﻿//-----------------------------------------------------------------------
// <copyright file="TestFileRuleSetProvider.cs" company="tuliprules.org">
//     Copyright (c) 2010 tuplelister, tuliprules.org.
// </copyright>
// <summary>This is the TestFileRuleSetProvider class.</summary>
//-----------------------------------------------------------------------

namespace Tulip.Rules.Tests
{
    using System;
    using System.Collections.Generic;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Rules;
    using System.Reflection;
    using System.Globalization;
    using System.IO;
    using System.Web;

    /// <summary>
    /// Tests the TestFileRuleSetProvider class.
    /// </summary>
    [TestClass]
    public class TestFileRuleSetProvider
    {
        /// <summary>
        /// Tests the FileRuleSetProvider for adding rule set dictionaries to a file.
        /// </summary>
        [TestMethod]
        public void AddRuleSetDictionaryToFile()
        {
            DeleteFile(@"c:\temp\TestRuleSets.xml");
            var fileRuleSetProvider = new FileRuleSetProvider(@"c:\temp\TestRuleSets.xml");
            fileRuleSetProvider.Add(CreateRulesetDictionary());
        }

        /// <summary>
        /// Tests the FileRuleSetProvider for adding rule sets to a file.
        /// </summary>
        [TestMethod]
        public void AddRuleSetToFile()
        {
            DeleteFile(@"c:\temp\TestRuleSets.xml");
            var fileRuleSetProvider = new FileRuleSetProvider(@"c:\temp\TestRuleSets.xml");

            var ruleSet = new RuleSet("PRD Running Example", 1, 0, 0, 0)
                              {
                                  Parser = new RuleSetParser()
                              };
            ruleSet.Parser.Parse(GetPresentationSyntax());

            fileRuleSetProvider.Add(ruleSet);
        }

        private static void DeleteFile(string path)
        {
            var fileInfo = new FileInfo(path);

            if (fileInfo.Exists)
            {
                fileInfo.Delete();
            }
        }

        /// <summary>
        /// Returns a rule set dictionary populated with four rulesets. 
        /// </summary>
        /// <returns>a rule set dictionary populated with four rulesets.</returns>
        private static RuleSetDictionary CreateRulesetDictionary()
        {
            var ruleSetDictionary = new RuleSetDictionary();
            var parser = new RuleSetParser();
            parser.Parse("This is a dummy ruleset.");

            var ruleSet1 = new RuleSet
                               {
                                   Name = "Ruleset 1",
                                   MajorVersion = 1,
                                   MinorVersion = 1,
                                   Build = 1,
                                   Revision = 1,
                                   Parser = parser
                               };

            var ruleSet2 = new RuleSet("Ruleset 2")
                               {
                                   MajorVersion = ruleSet1.MajorVersion,
                                   MinorVersion = ruleSet1.MinorVersion,
                                   Build = ruleSet1.Build,
                                   Revision = ruleSet1.Revision,
                                   Parser = ruleSet1.Parser
                               };

            var ruleSet3 = new RuleSet("Ruleset 3", 1, 0, 0, 0)
                               {
                                   Parser = parser
                               };

            var ruleSet4 = new RuleSet("Ruleset 4", new Version(1, 0, 0, 0))
                               {
                                   Parser = parser
                               };

            ruleSetDictionary.Add(ruleSet1.Name, ruleSet1);
            ruleSetDictionary[ruleSet2.Name] = ruleSet2;

            var kpv1 = new KeyValuePair<string, RuleSet>(ruleSet3.Name, ruleSet3);
            var kpv2 = new KeyValuePair<string, RuleSet>(ruleSet4.Name, ruleSet4);
            ruleSetDictionary.Add(kpv1);
            ruleSetDictionary.Add(kpv2);

            return ruleSetDictionary;
        }

        /// <summary>
        /// Returns an HTML-encoded version of a RIF-PRD resource (in presentation syntax)
        /// </summary>
        /// <returns></returns>
        private string GetPresentationSyntax()
        {
            var assembly = Assembly.GetExecutingAssembly();

            using (var stream = assembly.GetManifestResourceStream(Properties.Resources.RunningPrd))
            {
                if (stream == null)
                {
                    throw new RulesException();
                }

                var reader = new StreamReader(stream);
                return HttpUtility.HtmlEncode(reader.ReadToEnd());
            }
        }
    }
}
