﻿//-----------------------------------------------------------------------
// <copyright file="TestRuleSetDictionary.cs" company="tuliprules.org">
//     Copyright (c) 2010 tuplelister, tuliprules.org.
// </copyright>
// <summary>This is the TestRuleSetDictionary class.</summary>
//-----------------------------------------------------------------------

namespace Tulip.Rules.Tests
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using System.Xml;
    using System.Xml.Serialization;
    using System.IO;

    /// <summary>
    /// Test class for the RuleSetDictionary.
    /// </summary>
    [TestClass]
    public class TestRuleSetDictionary
    {
        /// <summary>
        /// Tests the TestRuleSetDictionary constructors.
        /// </summary>
        [TestMethod]
        public void TestRuleSetDictionaryConstruction()
        {
            var dictionary1 = new RuleSetDictionary();
            Assert.IsNotNull(dictionary1);
            var dictionary2 = new RuleSetDictionary(100);
            Assert.IsNotNull(dictionary2);
            var dictionary3 = new RuleSetDictionary(100, new RuleSetComparer());
            Assert.IsNotNull(dictionary3);
            var dictionary4 = new RuleSetDictionary(new RuleSetComparer());
            Assert.IsNotNull(dictionary4);
            var dictionary5 = new RuleSetDictionary(dictionary1);
            Assert.IsNotNull(dictionary5);
            var dictionary6 = new RuleSetDictionary(dictionary1, new RuleSetComparer());
            Assert.IsNotNull(dictionary6);
        }

        /// <summary>
        /// Tests various aspects of general use of the TestRuleSetDictionary class.
        /// </summary>
        [TestMethod]
        public void TestRuleSetDictionaryUse()
        {
            // Test the Add methods
            var ruleSetDictionary = CreateRulesetDictionary();

            // Check that rule set is writable.
            Assert.IsFalse(ruleSetDictionary.IsReadOnly, "Rule set dictionary reports that it is read-only.");

            // Test the Count property
            Assert.AreEqual(4, ruleSetDictionary.Count, "Unexpected number of rule sets forund in dictionary.");

            // Test the indexer
            Assert.IsNotNull(ruleSetDictionary["Ruleset 1"], "Cannot get ruleset 1 from dictionary using indexer.");
            Assert.IsNotNull(ruleSetDictionary["Ruleset 2"], "Cannot get ruleset 2 from dictionary using indexer.");
            Assert.IsNotNull(ruleSetDictionary["Ruleset 3"], "Cannot get ruleset 3 from dictionary using indexer.");
            Assert.IsNotNull(ruleSetDictionary["Ruleset 4"], "Cannot get ruleset 4 from dictionary using indexer.");

            // Test the version specifiers (to ensure we fully test the RuleSet)
            Assert.AreEqual(new Version(1, 0, 0, 0), ruleSetDictionary["Ruleset 1"].Version, "Unexpected version number reported for rule set 1.");
            Assert.AreEqual(new Version(1, 0, 0, 0), ruleSetDictionary["Ruleset 2"].Version, "Unexpected version number reported for rule set 2.");
            Assert.AreEqual(new Version(1, 0, 0, 0), ruleSetDictionary["Ruleset 3"].Version, "Unexpected version number reported for rule set 3.");
            Assert.AreEqual(new Version(1, 0, 0, 0), ruleSetDictionary["Ruleset 4"].Version, "Unexpected version number reported for rule set 4.");

            try
            {
                var ruleSet = ruleSetDictionary["Dummy ruleset"];
                Assert.IsNull(ruleSet, "Indexer unexpectedly returned a rule set for a dummy key.");
            }
            catch (KeyNotFoundException)
            {
                // Test passed.
            }

            try
            {
                var ruleSet = ruleSetDictionary[string.Empty];
                Assert.IsNull(ruleSet, "Indexer unexpectedly returned a rule set for an empty key.");
            }
            catch (KeyNotFoundException)
            {
                // Test passed.
            }

            var kpv1 = new KeyValuePair<string, RuleSet>("Ruleset 1", ruleSetDictionary["Ruleset 1"]);
            var kpv2 = new KeyValuePair<string, RuleSet>("Ruleset 2", ruleSetDictionary["Ruleset 2"]);
            var kpv3 = new KeyValuePair<string, RuleSet>("Ruleset 3", ruleSetDictionary["Ruleset 3"]);
            var kpv4 = new KeyValuePair<string, RuleSet>("Ruleset 4", ruleSetDictionary["Ruleset 4"]);

            var kpvBad1 = new KeyValuePair<string, RuleSet>("Dummy RuleSet", ruleSetDictionary["Ruleset 4"]);
            var kpvBad2 = new KeyValuePair<string, RuleSet>("Ruleset 4", ruleSetDictionary["Ruleset 1"]);
            var kpvBad3 = new KeyValuePair<string, RuleSet>(string.Empty, ruleSetDictionary["Ruleset 1"]);
            var kpvBad4 = new KeyValuePair<string, RuleSet>("Ruleset 1", null);

            // Test Contains method
            Assert.IsTrue(ruleSetDictionary.Contains(kpv1), "'Contains' method cannot find rule set 1.");
            Assert.IsTrue(ruleSetDictionary.Contains(kpv2), "'Contains' method cannot find rule set 2.");
            Assert.IsTrue(ruleSetDictionary.Contains(kpv3), "'Contains' method cannot find rule set 3.");
            Assert.IsTrue(ruleSetDictionary.Contains(kpv4), "'Contains' method cannot find rule set 4.");
            Assert.IsTrue(ruleSetDictionary.Contains(kpv4), "'Contains' method cannot find rule set 4.");

            Assert.IsFalse(ruleSetDictionary.Contains(kpvBad1), "'Contains' method found non-existent dummy rule set.");
            Assert.IsFalse(ruleSetDictionary.Contains(kpvBad2), "'Contains' method found rule set where key does not match value.");
            Assert.IsFalse(ruleSetDictionary.Contains(kpvBad3), "'Contains' method found rule set given an empty key.");
            Assert.IsFalse(ruleSetDictionary.Contains(kpvBad4), "'Contains' method found a rule set when given a null value.");

            // Test ContainsKey method
            Assert.IsTrue(ruleSetDictionary.ContainsKey("Ruleset 1"), "'ContainsKey' method cannot find rule set 1.");
            Assert.IsTrue(ruleSetDictionary.ContainsKey("Ruleset 2"), "'ContainsKey' method cannot find rule set 2.");
            Assert.IsTrue(ruleSetDictionary.ContainsKey("Ruleset 3"), "'ContainsKey' method cannot find rule set 3.");
            Assert.IsTrue(ruleSetDictionary.ContainsKey("Ruleset 4"), "'ContainsKey' method cannot find rule set 4.");

            Assert.IsFalse(ruleSetDictionary.ContainsKey("Dummy RuleSet"), "'ContainsKey' method found non-existent dummy rule set.");
            Assert.IsFalse(ruleSetDictionary.ContainsKey(string.Empty), "'ContainsKey' method found rule set where key does not match value.");

            // Test TryGetValue method
            RuleSet ruleSet1;
            Assert.IsTrue(ruleSetDictionary.TryGetValue("Ruleset 1", out ruleSet1), "'TryGetValue' reports it did not find rule set 1.");
            Assert.IsNotNull(ruleSet1, "'TryGetValue' did not return a rule set.");
            Assert.AreSame(ruleSetDictionary["Ruleset 1"], ruleSet1, "'TryGetValue' returned unexpected rule set.");
            RuleSet ruleSet2;
            Assert.IsTrue(ruleSetDictionary.TryGetValue("Ruleset 2", out ruleSet2), "'TryGetValue' reports it did not find rule set 2.");
            Assert.IsNotNull(ruleSet2, "'TryGetValue' did not return a rule set.");
            Assert.AreSame(ruleSetDictionary["Ruleset 2"], ruleSet2, "'TryGetValue' returned unexpected rule set.");
            RuleSet ruleSet3;
            Assert.IsTrue(ruleSetDictionary.TryGetValue("Ruleset 3", out ruleSet3), "'TryGetValue' reports it did not find rule set 3.");
            Assert.IsNotNull(ruleSet3, "'TryGetValue' did not return a rule set.");
            Assert.AreSame(ruleSetDictionary["Ruleset 3"], ruleSet3, "'TryGetValue' returned unexpected rule set.");
            RuleSet ruleSet4;
            Assert.IsTrue(ruleSetDictionary.TryGetValue("Ruleset 4", out ruleSet4), "'TryGetValue' reports it did not find rule set 4.");
            Assert.IsNotNull(ruleSet4, "'TryGetValue' did not return a rule set.");
            Assert.AreSame(ruleSetDictionary["Ruleset 4"], ruleSet4, "'TryGetValue' returned unexpected rule set.");

            RuleSet ruleSet5;
            Assert.IsFalse(ruleSetDictionary.TryGetValue("Ruleset 5", out ruleSet5), "'TryGetValue' reports it found non-existent rule set 5.");
            Assert.IsNull(ruleSet5, "'TryGetValue' returned a rule set unexpectedly.");

            // Test CopyTo method
            var kpvArray1 = new KeyValuePair<string, RuleSet>[ruleSetDictionary.Count];
            ruleSetDictionary.CopyTo(kpvArray1, 0);
            Assert.AreEqual(4, kpvArray1.Length, "Incorrect number of key value pairs copied to array.");

            var kpvArray2 = new KeyValuePair<string, RuleSet>[ruleSetDictionary.Count - 2];
            ruleSetDictionary.CopyTo(kpvArray2, 2);
            Assert.AreEqual(2, kpvArray2.Length, "Incorrect number of key value pairs copied to array.");

            // Test Remove method
            Assert.IsFalse(ruleSetDictionary.Remove(kpvBad1), "'Remove' reported success when removing a non-existent rule set.");
            Assert.AreEqual(4, ruleSetDictionary.Count, "Dictionary reports an unexpected number of rule sets.");
            Assert.IsFalse(ruleSetDictionary.Remove("Dummy RuleSet"), "'Remove' reported success when removing a non-existent rule set.");
            Assert.AreEqual(4, ruleSetDictionary.Count, "Dictionary reports an unexpected number of rule sets.");

            Assert.IsTrue(ruleSetDictionary.Remove(kpv3), "'Remove' reported failure to remove rule set 3.");
            Assert.AreEqual(3, ruleSetDictionary.Count, "Dictionary reports an unexpected number of rule sets.");
            Assert.IsTrue(ruleSetDictionary.Remove("Ruleset 4"), "'Remove' reported failure to remove rule set 4.");
            Assert.AreEqual(2, ruleSetDictionary.Count, "Dictionary reports an unexpected number of rule sets.");

            // Test Clear method
            ruleSetDictionary.Clear();
            Assert.AreEqual(0, ruleSetDictionary.Count, "Dictionary reports an unexpected number of rule sets after clearing.");
        }

        /// <summary>
        /// Tests enumeration of the TestRuleSetDictionary class.
        /// </summary>
        [TestMethod]
        public void TestRuleSetDictionaryEnumeration()
        {
            // Populate the rules dictionary
            var ruleSetDictionary = CreateRulesetDictionary();

            var idx = 0;
            foreach (var ruleSet in ruleSetDictionary)
            {
                idx++;
                Assert.IsNotNull(ruleSet, "Enumerator failed to return a rule set.");
            }

            Assert.AreEqual(4, idx, "Enumerator found unexpected number of rule sets.");

            idx = 0;
            var enumerator = ((IEnumerable)ruleSetDictionary).GetEnumerator();
            Assert.IsNotNull(enumerator, "Failed to get an IEnumerable enumerator from GetEnumerator.");
            while (enumerator.MoveNext())
            {
                idx++;
                Assert.IsNotNull(enumerator.Current, "Enumerator failed to return a rule set.");
            }

            Assert.AreEqual(4, idx, "Enumerator found unexpected number of rule sets.");

            idx = 0;
            foreach (var key in ruleSetDictionary.Keys)
            {
                idx++;
                Assert.IsFalse(string.IsNullOrEmpty(key), "Enumerator key collection contains an empty or null key.");
            }

            Assert.AreEqual(4, idx, "Key enumerator found unexpected number of keys.");

            idx = 0;
            foreach (var value in ruleSetDictionary.Values)
            {
                idx++;
                Assert.IsNotNull(value, "Enumerator key collection contains a null value.");
            }

            Assert.AreEqual(4, idx, "Key enumerator found unexpected number of values.");
        }


        /// <summary>
        /// Tests serialization and deserialization of the RuleSetDictionary class.
        /// </summary>
        [TestMethod]
        public void TestRuleSetDictionarySerialization()
        {
            var ruleSetDictionary = CreateRulesetDictionary();

            var serializer = new XmlSerializer(typeof(RuleSetDictionary));

            using (var ms = new MemoryStream())
            {
                serializer.Serialize(ms, ruleSetDictionary);
                var xmlDocument = new XmlDocument();
                ms.Seek(0, SeekOrigin.Begin);
                xmlDocument.Load(ms);

                // Deserialize back to RuleSetDictionary
                using (var msOut = new MemoryStream())
                {
                    xmlDocument.Save(msOut);
                    msOut.Seek(0, SeekOrigin.Begin);
                    var ruleSetDictionaryOut = (RuleSetDictionary)serializer.Deserialize(msOut);

                    Assert.AreEqual(ruleSetDictionary.Count, ruleSetDictionaryOut.Count, "The deserialized rule set dictionary does not have the same count as the original.");

                    foreach (var ruleSetOut in ruleSetDictionaryOut)
                    {
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].MajorVersion, ruleSetOut.Value.MajorVersion, "The deserialized rule set does not have the same name as the original.");
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].MajorVersion, ruleSetOut.Value.MajorVersion, "The deserialized rule set does not have the same major version as the original.");
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].MinorVersion, ruleSetOut.Value.MinorVersion, "The deserialized rule set does not have the same minor version as the original.");
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].Build, ruleSetOut.Value.Build, "The deserialized rule set does not have the same build as the original.");
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].Revision, ruleSetOut.Value.Revision, "The deserialized rule set does not have the same revision as the original.");
                        Assert.AreEqual(ruleSetDictionary[ruleSetOut.Key].Presentation, ruleSetOut.Value.Presentation, "The deserialized rule set does not have the same content as the original.");
                    }
                }
            }
        }

        /// <summary>
        /// Returns a rule set dictionary populated with four rulesets. 
        /// </summary>
        /// <returns>a rule set dictionary populated with four rulesets.</returns>
        private static RuleSetDictionary CreateRulesetDictionary()
        {
            var ruleSetDictionary = new RuleSetDictionary();
            var parser = new RuleSetParser();
            parser.Parse("This is a dummy ruleset.");

            var ruleSet1 = new RuleSet
            {
                Name = "Ruleset 1",
                MajorVersion = 1,
                MinorVersion = 0,
                Build = 0,
                Revision = 0,
                Parser = parser
            };

            var ruleSet2 = new RuleSet("Ruleset 2")
            {
                MajorVersion = ruleSet1.MajorVersion,
                MinorVersion = ruleSet1.MinorVersion,
                Build = ruleSet1.Build,
                Revision = ruleSet1.Revision,
                Parser = ruleSet1.Parser
            };

            var ruleSet3 = new RuleSet("Ruleset 3", 1, 0, 0, 0)
            {
                Parser = parser
            };

            var ruleSet4 = new RuleSet("Ruleset 4", new Version(1, 0, 0, 0))
            {
                Parser = parser
            };

            ruleSetDictionary.Add(ruleSet1.Name, ruleSet1);
            ruleSetDictionary[ruleSet2.Name] = ruleSet2;

            var kpv1 = new KeyValuePair<string, RuleSet>(ruleSet3.Name, ruleSet3);
            var kpv2 = new KeyValuePair<string, RuleSet>(ruleSet4.Name, ruleSet4);
            ruleSetDictionary.Add(kpv1);
            ruleSetDictionary.Add(kpv2);

            return ruleSetDictionary;
        }

        /// <summary>
        /// String comparer used during testing.
        /// </summary>
        private class RuleSetComparer : StringComparer
        {
            /// <summary>
            /// Compares two strings.
            /// </summary>
            /// <param name="string1">The first string.</param>
            /// <param name="string2">The second string.</param>
            /// <returns>)0, if the string are equal; otherwise a non-zero integer.</returns>
            public override int Compare(string string1, string string2)
            {
                if (string1 == null)
                {
                    return -1;
                }

                return string1.CompareTo(string2);
            }

            /// <summary>
            /// Compares two strings for equality.
            /// </summary>
            /// <param name="string1">The first string.</param>
            /// <param name="string2">The second string.</param>
            /// <returns>True, if the strings are equal; otherwise false.</returns>
            public override bool Equals(string string1, string string2)
            {
                return string1 != null && string1.Equals(string2);
            }

            /// <summary>
            /// Returns a hash code foe the string provide.
            /// </summary>
            /// <param name="value">A string value.</param>
            /// <returns>A hash code for the provided string.</returns>
            public override int GetHashCode(string value)
            {
                return string.IsNullOrWhiteSpace(value) ? 0 : value.GetHashCode();
            }
        }
    }
}
